/*
 * Copyright 2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.callbackparams.internal.annotation;

import java.lang.reflect.Array;
import org.callbackparams.internal.CallbackContext;
import java.util.ArrayList;
import java.util.Collection;
import java.util.IdentityHashMap;
import java.util.List;
import java.util.Map;
import java.util.ListIterator;

/**
 * @author Henrik Kaipe
 */
class CallbackContextBridge
extends org.callbackparams.internal.CallbackContextBridge {

    private final static ThreadLocal<Map<Object,Object>> nestedArrays =
            new ThreadLocal<Map<Object, Object>>() {
        @Override
        protected Map<Object, Object> initialValue() {
            return new IdentityHashMap<Object, Object>();
        }
    };

    CallbackContextBridge(
            ClassLoader bridgedLoader,
            CallbackContext bridgedContext) {
        super(bridgedLoader, bridgedContext);
    }

    /**
     * This method can only handle bridge record-values that are Enum constants.
     * Other values will be left as-is.
     */
    public List getCallbackRecord() {
        List list = bridgedContext.getCallbackRecord();
        for (ListIterator li = list.listIterator() ; li.hasNext() ;) {
            final Object nextValue = li.next();
            if (nextValue instanceof Enum) {
                li.set(findEquivalentEnum(nextValue, localLoader));
            }
        }
        return list;
    }

    /**
     * This method can only handle records that consists exclusivly of
     * Enum-constants.
     */
    @Override
    public void setCallbackRecord(Collection callbackRecord) {
        List records = new ArrayList(callbackRecord.size());
        for (Object r : callbackRecord) {
            records.add(findEquivalent(r, bridgedLoader));
        }
        bridgedContext.setCallbackRecord(records);
    }

    static Object findEquivalent(Object o, ClassLoader searchLoader) {
        if (null == o) {
            return null;
        }

        Class<?> c = o.getClass();
        if (c.isArray()) {
            return createEquivalentArray(o, searchLoader);
        } else {
            return findEquivalentEnum(o, searchLoader);
        }
    }

    private static Object createEquivalentArray(
            Object o, ClassLoader searchLoader) {
        Class<?> componentType = o.getClass().getComponentType();
        if (componentType.isPrimitive()) {
            return o;

        } else if (nestedArrays.get().containsKey(o)) {
            return nestedArrays.get().get(o);

        } else {
            Object[] array = (Object[])o;
            Object[] eqArray = (Object[]) Array.newInstance(
                    findEquivalentClass(componentType, searchLoader),
                    array.length);
            nestedArrays.get().put(array, eqArray);
            for (int i = 0 ; i < eqArray.length ; ++i) {
                eqArray[i] = findEquivalent(array[i], searchLoader);
            }
            return nestedArrays.get().remove(array);
        }
    }

    private static Class<?> findEquivalentClass(
            Class<?> c, ClassLoader searchLoader) {
        if (false == c.isArray()) {
            try {
                return searchLoader.loadClass(c.getName());
            } catch (ClassNotFoundException ex) {
                throw new Error("Cannot find equivalent of " + c
                    + " on class-loader " + searchLoader, ex);
            }

        } else {
            Class<?> componentType = c.getComponentType();
            if (componentType.isPrimitive()) {
                return c;
            } else {
                return Array.newInstance(findEquivalentClass(
                        componentType, searchLoader), 0).getClass();
            }
        }
    }

    /**
     * @param knownEnumConstant a callback value, expected to be an Enum constant
     * @param searchLoader a classloader
     * @return the equivalent of knownEnumConstant from
     *         the specified class-loader searchLoader
     */
    private static Object findEquivalentEnum(
            Object knownEnumConstant, ClassLoader searchLoader) {
        if (null == knownEnumConstant) {
            return null;
        }

        Class<?> eqClass = findEquivalentClass(
                knownEnumConstant.getClass(), searchLoader);
        if (knownEnumConstant.getClass() == eqClass) {
            return knownEnumConstant;
        } else {
            Enum<?> enm = (Enum<?>) knownEnumConstant;
            eqClass = findEquivalentClass(
                    enm.getDeclaringClass(), searchLoader);
            return Enum.valueOf(eqClass.asSubclass(Enum.class), enm.name());
        }
    }
}
