/*
 * Copyright 2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.callbackparams.junit3;

import java.util.Collection;
import junit.framework.TestCase;
import junit.framework.TestSuite;
import org.callbackparams.internal.CallbackContextFactory;
import org.callbackparams.support.ThreadContext;

/**
 * Works as a wrapper around a callback testrun and makes sure the testrun will
 * be properly named and have the right callback-record in context when the
 * test-instance is created.<br/>
 * Please not that ordinary JUnit-3.8.x testruns are named on the format
 * &lt;test-method name&gt;(&lt;test-class name&gt;)
 * whereas callback testruns have names on the format
 * &lt;test-method name&gt;[&lt;callback record-values&gt;](&lt;test-class name&gt;).
 *
 * This class is used by TestInstanceFactory to wrap the test-instances.
 * However, this class is never used directly but renamed through byte-code
 * modification before usage, so that the class-name will the same as the name
 * of the test-class. - Otherwise some report tools (such as surefire report)
 * would forget about the original test-class name and instead use the name
 * "CallbackTestrunWrapper".
 *
 * @author Henrik Kaipe
 */
class CallbackTestrunWrapper extends TestCase implements ThreadContext.Testrun {

    private final Class rebytedTestClass;
    private final String noargTestMethodName;
    private final String actualTestMethodName;
    private final Collection testRunCallbackRecord;

    public CallbackTestrunWrapper(Class rebytedTestClass,
            String noargTestMethodName, String actualTestMethodName,
            Collection testRunCallbackRecord) {
        this.rebytedTestClass = rebytedTestClass;
        this.noargTestMethodName = noargTestMethodName;
        this.actualTestMethodName = actualTestMethodName;
        this.testRunCallbackRecord = testRunCallbackRecord;
    }

    /**
     * This method is invoked by the test-runners' testresult instance. It makes
     * sure the proper callback record is setup in context then the
     * invocation is simply forwarded to {@link #wrappedTestCase}.
     *
     * @throws java.lang.Throwable test-failure or test-error
     */
    public void runBare() throws Throwable {
        ThreadContext.runWithCtxClassLoader(
                rebytedTestClass.getClassLoader(), this);
    }

    /**
     * @return &lt;test-method name&gt;[&lt;callback record-values&gt;]
     */
    public String getName() {
        return actualTestMethodName + testRunCallbackRecord;
    }

    public void executeTestrun() throws Throwable {
        setupCallbackRecord();
        createTest().runBare();
    }

    private void setupCallbackRecord() throws Exception {
        CallbackContextFactory.forClassLoader(rebytedTestClass.getClassLoader())
                .setCallbackRecord(testRunCallbackRecord);
    }

    private TestCase createTest() {
        return (TestCase) TestSuite
                .createTest(rebytedTestClass, noargTestMethodName);
    }
}
