/*
 * Copyright 2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.callbackparams.junit4;

import org.callbackparams.junit4.support.CallbackTestRunIterator;
import java.lang.annotation.Annotation;
import java.lang.annotation.Documented;
import java.lang.annotation.ElementType;
import java.lang.annotation.Inherited;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import org.callbackparams.combine.reflect.CallbackRecordsFactory;
import org.callbackparams.junit4.support.ExperimentalUtils;
import org.callbackparams.junit4.support.JUnit4CallbackParamsFilteredExtension;
import org.callbackparams.junit4.support.JUnit4WithCallbackParamsExtension;
import org.callbackparams.junit4.support.NeedsCallbackArguments;
import org.callbackparams.support.ClassConstantModifyingVisitor;
import org.callbackparams.support.ClassWrapper;
import org.callbackparams.support.TemplateBasedClassBuilder;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.Description;
import org.junit.runner.notification.RunNotifier;
import org.junit.runners.BlockJUnit4ClassRunner;
import org.junit.runners.ParentRunner;
import org.junit.runners.model.FrameworkMethod;
import org.junit.runners.model.InitializationError;

/**
 * An experimental alternative to {@link CallbackParamsRunner} that can only be
 * used under JUnit version 4.5 or later. Unlike
 * {@link CallbackParamsRunner} it does
 * not support callback-injection (i.e. the annotation &#64;CallbackField)
 * and the JUnit-3.x API. It has its own nested annotation
 * {@link ExperimentalCallbackParamsRunner.WrappedRunner @WrappedRunner} and can
 * only accept runner implementations that subclass
 * <code>BlockJUnit4ClassRunner</code>. (This is a major disadvantage because as
 * of today there are only a few third-party Runner implementations that fulfill
 * this criteria.) The only advantage is that it does not modify the test-class
 * bytecode and therefore does not need to reload any (existing) classes through
 * the new classloader.
 * <br/><br/>
 * The ideology difference between this class and
 * {@link CallbackParamsRunner} is
 * that the latter enforces the functionality by making bytecode-modifications
 * on the test-class, thus gaining tremendous flexibility but having to reload
 * many classes through the rebyted test-class classloader. This class, on the
 * other hand, tries to enforce the same functionality by tempering with the
 * byte-code of the Runner implementation (i.e.
 * <code>BlockJUnit4ClassRunner</code> or a subclass thereof), which seems to be
 * a lot harder and less flexible, so that only JUnit-4.5+ and
 * <code>BlockJUnit4ClassRunner</code> implementations are supported.
 * <br/>
 * If the features of this class will stay desirable in the future they will
 * probably be incorporated within <code>CallbackParamsRunner</code> so that
 * the desirable ideology will be determined at runtime, i.e. the
 * functionality of this class is kicked-in only when applyable.
 *
 * @author Henrik Kaipe
 */
public class ExperimentalCallbackParamsRunner
extends ParentRunner<BlockJUnit4ClassRunner> {
    
//    /**
//     * By default ExperimentalCallbackParamsRunner wraps an extended version of the class
//     * BlockJUnit4ClassRunner to run the tests. The extension overrides
//     * the method-invocation on the test-instance so that callback parameters
//     * can be passed as arguments.
//     * By using this annotation it is possible to wrap a class-runner other than
//     * BlockJUnit4ClassRunner. However, it is important that this runner is an
//     * extension of BlockJUnit4ClassRunner and that it does not mess with the
//     * method-invocations, or it would probably conflict with
//     * ExperimentalCallbackParamsRunner's manipulation of the method invocations.
//     * I.e. suitable runner-classes should stick to manipulating the
//     * test-instances and possibly add some additional pre- or post-tasks to
//     * the test. This is how most extensions to BlockJUnit4ClassRunner act so
//     * it will usually work out fine.
//     * Examples of popular runner-classes that can be wrapped by
//     * ExperimentalCallbackParamsRunner are JMock (version 2.6+) and SpringJUnit4ClassRunner
//     * (version 3.0+).
//     */
    /**
     * This nested annotation is to {@link ExperimentalCallbackParamsRunner}
     * what top-level annotation
     * {@link org.callbackparams.junit4.WrappedRunner @WrappedRunner} is to
     * {@link CallbackParamsRunner}.
     * <br/>
     * The reason for having a separate annotation for
     * {@link ExperimentalCallbackParamsRunner} is that it can only wrap
     * a runner implementation class that is a subclass of
     * <code>BlockJUnit4ClassRunner</code>.
     */
    @Retention(RetentionPolicy.RUNTIME)
    @Target(ElementType.TYPE)
    @Inherited
    @Documented
    public @interface WrappedRunner {
        Class<? extends BlockJUnit4ClassRunner> value();
    }
    
    /**
     * Internally used interface for the possibly byte-code manipulating
     * factories that are used for creating callback-versions of the test-runner
     * class specified by the {@link WrappedRunner} annotation (or
     * BlockJUnit4ClassRunner instances if no wrapped-runner has been specified).
     */
    private interface TestRunnerFactory {
        BlockJUnit4ClassRunner newInstance();
    }
    
    private final List<BlockJUnit4ClassRunner> children;
    
    public ExperimentalCallbackParamsRunner(Class testClass)
    throws InitializationError {
        super(testClass);
        this.children = createChildren();
    }
    
    private List<BlockJUnit4ClassRunner> createChildren()
    throws InitializationError {
        final Collection<Method> callbackTestMethods = getCallbackTestMethods();
        final CallbackTestRunIterator callbackTestRuns =
                new CallbackTestRunIterator(
                getCallbackRecords(), callbackTestMethods);
        final List<BlockJUnit4ClassRunner> childList =
                new ArrayList(callbackTestRuns.totalSize() + 1);
        
        if (callbackTestMethods.size()
                < getTestClass().getAnnotatedMethods(Test.class).size()) {
            childList.add(newTestRunnerFactory(
                    JUnit4CallbackParamsFilteredExtension.class).newInstance());
        }
        
        final TestRunnerFactory callbackParamsExtendedRunnerFactory =
                newTestRunnerFactory(JUnit4WithCallbackParamsExtension.class);
        while (callbackTestRuns.next()) {
            final BlockJUnit4ClassRunner test =
                    callbackParamsExtendedRunnerFactory.newInstance();
            ((NeedsCallbackArguments)test).setupCallbackArguments(
                    callbackTestRuns.getCurrentCallbackTestMethod(),
                    callbackTestRuns.getCurrentCallbackRecord());
            childList.add(test);
        }
        
        return Collections.unmodifiableList(childList);
    }
            
    private Collection getCallbackRecords() {
        return CallbackRecordsFactory.getInstance()
                .collectCallbackRecordsReflectively(
                getTestClass().getJavaClass());
    }
    
    /**
     * @return all test methods in case there are callback-params methods among
     * the befores or afters; otherwise only the test-methods that take
     * callback-params
     */
    private Collection<Method> getCallbackTestMethods() {
        if (getCallbackParamsMethods(Before.class).isEmpty()
                && getCallbackParamsMethods(After.class).isEmpty()) {
            return getCallbackParamsMethods(Test.class);
            
        } else {
            final Collection<FrameworkMethod> fmTestMethods =
                    getTestClass().getAnnotatedMethods(Test.class);
            final Collection<Method> testMethods =
                    new ArrayList<Method>(fmTestMethods.size());
            for (final FrameworkMethod fmEach : fmTestMethods) {
                testMethods.add(fmEach.getMethod());
            }
            return testMethods;
        }
    }
    
    private Collection<Method> getCallbackParamsMethods(
            final Class<? extends Annotation> annotation) {
        final Collection<Method> methods = new ArrayList<Method>();
        
        for (final FrameworkMethod fm
                : getTestClass().getAnnotatedMethods(annotation)) {
            final Method m = fm.getMethod();
            if (ExperimentalUtils.isCallbackParamsMethod(m)) {
                methods.add(m);
            }
        }
        
        return methods;
    }

    @Override
    protected List<BlockJUnit4ClassRunner> getChildren() {
        return children;
    }

    @Override
    protected Description describeChild(BlockJUnit4ClassRunner child) {
        return child.getDescription();
    }

    @Override
    protected void runChild(
            BlockJUnit4ClassRunner child, RunNotifier notifier) {
        child.run(notifier);
    }
    
    /**
     * Creates a factory that will create (possibly rebyted) instances of the
     * specified extensionTemplateClass.
     */
    private TestRunnerFactory newTestRunnerFactory(
            Class<? extends BlockJUnit4ClassRunner> extensionTemplateClass) {
        
        /****************************************************************
         * Get hold of the annotation that specifies wrapped test-runner
         ****************************************************************/
        WrappedRunner wr = getTestClass().getJavaClass()
                .getAnnotation(WrappedRunner.class);
        
        final ClassWrapper testRunnerWrapper;
        if (null == wr || BlockJUnit4ClassRunner.class == wr.value()) {
            /************************************************************
             * No alternative test-runner was specified.
             * The template-class can be used as-is.
             ************************************************************/ 
            testRunnerWrapper = new ClassWrapper(extensionTemplateClass);
            
        } else {
            /********************************************************
             * A special test-runner class has been specified.
             * Byte-code manipulation is needed!
             ********************************************************/
            ClassLoader testClassClassLoader =
                    getTestClass().getJavaClass().getClassLoader();
            ClassConstantModifyingVisitor modifiesSuperClass =
                    new ClassConstantModifyingVisitor(
                    BlockJUnit4ClassRunner.class, wr.value());
            
            testRunnerWrapper = TemplateBasedClassBuilder
                    .newBuilder(testClassClassLoader, extensionTemplateClass)
                    .applyVisitor(modifiesSuperClass)
                    .wrapClass();
        }
        
        /**********************************************************************
         * The child test-runner class is retrieved. Now we finally create
         * and return the factory that will create the test-runner instances.
         **********************************************************************/
        return new TestRunnerFactory() {
            public BlockJUnit4ClassRunner newInstance() {
                return (BlockJUnit4ClassRunner) testRunnerWrapper
                        .newInstance(getTestClass().getJavaClass());
            }
        };
    }
}
