/*
 * Copyright 2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.callbackparams.junit4.support;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.util.List;
import java.util.ListIterator;
import org.callbackparams.internal.template.CallbackRecord;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.junit.runners.BlockJUnit4ClassRunner;
import org.junit.runners.model.FrameworkMethod;
import org.junit.runners.model.InitializationError;

/**
 * @author Henrik Kaipe
 */
public class JUnit4WithCallbackParamsExtension extends BlockJUnit4ClassRunner
implements NeedsCallbackArguments {
    
    private static final Class[] testPrepAnnotations =
            {Before.class, After.class};
    
    public JUnit4WithCallbackParamsExtension(Class<?> klass)
    throws InitializationError {
        super(klass);
    }

    public void setupCallbackArguments(
            CallbackMethod callbackTestMethod,
            CallbackRecord testRunCallbackRecord) {

        /************************************
         * Setup single callback test-method
         ************************************/
        List<FrameworkMethod> testMethods =
                getTestClass().getAnnotatedMethods(Test.class);
        testMethods.clear();
        final CallbackFrameworkMethod callbackFrameworkMethod =
                new CallbackFrameworkMethod(callbackTestMethod);
        callbackFrameworkMethod.setArgumentRecord(testRunCallbackRecord);
        testMethods.add(callbackFrameworkMethod);

        /****************************************************
         * Setup method arguments for all befores and afters
         ****************************************************/
        for (final Class<? extends Annotation> c : testPrepAnnotations) {
            for (FrameworkMethod fm : getTestClass().getAnnotatedMethods(c)) {
                if (fm instanceof CallbackFrameworkMethod) {
                    ((CallbackFrameworkMethod) fm)
                            .setArgumentRecord(testRunCallbackRecord);
                }
            }
        }
    }

    /**
     * This method is overridden and now does nothing. A new single
     * FrameworkMethod instance will later replace all test-methods by
     * {@link #setupCallbackArguments(CallbackEngine)} and that will also be
     * the only FrameworkMethod to be run by this Runner instance.
     */
    @Override
    protected void validateTestMethods(List<Throwable> errors) {
        /* No operation */
    }

    /**
     * FrameworkMethod instances for befores and afters that take callback-
     * params are replaced with modified versions that do not cause validation
     * failures because of the "NoArgs" condition.
     */
    @Override
    protected void collectInitializationErrors(List<Throwable> errors) {
        for (final Class<? extends Annotation> c : testPrepAnnotations) {
            modifyCallbackFrameworkMethods(c);
        }
        
        super.collectInitializationErrors(errors);
    }

    private void modifyCallbackFrameworkMethods(
            Class<? extends Annotation> annotation) {
        final ListIterator<FrameworkMethod> li =
                getTestClass().getAnnotatedMethods(annotation).listIterator();
        while (li.hasNext()) {
            final Method currentMethod = li.next().getMethod();
            
            if (ExperimentalUtils.isCallbackParamsMethod(currentMethod)) {
                /* Replace with "our" version of FrameworkMethod ... */
                li.set(new CallbackFrameworkMethod(
                        new CallbackMethod(currentMethod)));
            }
        }
    }
}
