/*
 * Copyright 2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.callbackparams.sandbox;

import java.lang.annotation.Annotation;
import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.Collection;
import org.callbackparams.annotation.util.EnumConstantDescriptor;
import org.callbackparams.factory.CallbackFactory;

/**
 * When used as a callback-interface, this single-method interface leverages
 * bleeding-edge framework features that allows callback-value declaration
 * through annotations.
 * <br/><br/>
 * The example use-case looks something like this ...
 * <pre>
 * &#64;Retention(RetentionPolicy.RUNTIME)
 * &#64;interface Throws {
 *   Class&lt;? extends Exception&gt; value();
 * }
 *
 * &#64;RunWith({@link org.callbackparams.junit4.CallbackParamsRunner})
 * public class MyTest {
 *
 *   Set&lt;Class&lt;?&gt;&gt; expectedExceptions = new HashSet();
 *
 *   &#64;Before
 *   public void collectExpectedExceptions(<b>Collector&lt;Throws&gt; exceptionCollector</b>) { // <b>(1)</b>
 *     exceptionCollector.collect(expectedExceptions);
 *   }
 *
 *   &#64;Test
 *   public void runTest(<b>DummyCallbackInterface callback</b>) { // <b>(2)</b>
 *
 *     try {
 *       // ...
 *       // Test-case execution functionality goes here
 *       // ...
 *
 *       assertTrue("Expected exception(s): " + expectedExceptions,
 *           expectedExceptions.isEmpty()
 *
 *     } catch (Exception x) {
 *       assertTrue("Expected exception(s) " + expectedExceptions + " but was " + x.getClass(),
 *           expectedExceptions.contains(x.getClass());
 *     }
 *   }
 *
 *   enum MyEnum implements DummyCallbackInterface {
 *     &#64;Throws(IllegalArgumentException.class) EMPTY(""),
 *     &#64;Throws(IllegalArgumentException.class) BLANK("   "),
 *     &#64;Throws(NullPointerException.class)     NULL(null),
 *     FOO("foo"), BAR("bar");
 *
 *     // ...
 *     // enum-body goes here
 *     // ...
 *   }
 * }
 * </pre>
 * The above test-class uses two callback-interfaces -
 * <code>DummyCallbackInterface</code> (2)
 * and <code>Collector</code> (1).
 * <br/>
 * Five callback-values are provided by the enum <code>MyEnum</code> and since
 * the enum implements <code>DummyCallbackInterface</code> by itself, there will
 * be exactly one callback-value for <code>DummyCallbackInterface</code> per
 * test-run. On the other hand -
 * <code>Collector</code> is not implemented which implies there
 * won't be any <code>Collector</code> callback-values for any of the test-runs.
 * <br/>
 * However, what really happens is that three of the <code>MyEnum</code>
 * constants (EMPTY, BLANK & NULL) will actually produce one callback-value
 * each, which {@link #collect(Collection)} implementation
 * will add the class-instance of their
 * respective <code>Throws</code> annotation to the argument collection. The
 * explanation to this magic is nested within the interface
 * <code>Collector</code>, where there is a class-declaration that will work as
 * a backup <code>Collector</code> factory for every enum constant that neither
 * implements <code>Collector</code> nor provides any implementation
 * through {@link org.callbackparams.CallbackFactory#getCallback()}. The
 * <code>Collector</code> factory will check its callback-declaration for
 * a generic parameter annotation-class (i.e. <code>Throws</code> in the
 * above case at (1)) and, if an enum-constant is annotated with that very
 * annotation class, the return-value(s) of the annotation's
 * <code>value()</code> method
 * will be added to the argument collection in the resulting
 * {@link #collect(Collection)} implementation.
 * <br/>
 * The nested
 * factory-class takes advantage of a bleeding-edge (i.e. not yet stable) API,
 * which primary purpose is to provide shortcuts
 * frequent for cross-cutting concerns that through annotations can
 * eliminate considerable amounts of copy-paste code.
 *
 * @author Henrik Kaipe
 */
public interface Collector<A extends Annotation> {

    /**
     * Implementations are expected to add objects to the collection.
     */
    <T> void collect(Collection<T> collection);

    /**
     * Works as a backup {@link Collector} factory for every enum
     * constant that neither
     * implements {@link Collector} nor provides any implementation
     * through {@link org.callbackparams.CallbackFactory#getCallback()}.
     */
    class CollectorFactory implements CallbackFactory {

        private final Class<? extends Annotation> specifiedAnnotation;
        private final Method valueMethod;

        private CollectorFactory(Type genericDecl) {
            if (genericDecl instanceof Class) {
                this.specifiedAnnotation = null;
                this.valueMethod = null;
            } else {
                Type a = ((ParameterizedType)genericDecl)
                        .getActualTypeArguments()[0];
                if (false == a instanceof Class) {
                    throw new AssertionError("Unsupported typeparameter: " + a);
                } else {
                    this.specifiedAnnotation =
                            ((Class)a).asSubclass(Annotation.class);
                    try {
                        this.valueMethod = this.specifiedAnnotation
                                .getMethod("value");
                        this.valueMethod.setAccessible(true);
                    } catch (NoSuchMethodException ex) {
                        throw new Error(ex);
                    }
                }
            }
        }

        public CollectorFactory(Method m, int paramIndex) {
            this(m.getGenericParameterTypes()[paramIndex]);
        }

        public CollectorFactory(Field f) {
            this(f.getGenericType());
        }

        public Object asCallback(Object callbackParameter) {
            if (null == this.valueMethod
                    || false == callbackParameter instanceof Enum) {
                return null;
            }
            Annotation annotation =
                    new EnumConstantDescriptor((Enum<?>) callbackParameter)
                    .getAnnotation(this.specifiedAnnotation);
            if (null == annotation) {
                return null;
            }
            try {
                final Object value = valueMethod.invoke(annotation);
                if (valueMethod.getReturnType().isArray()) {
                    final int length = Array.getLength(value);
                    return new Collector() {
                        public void collect(Collection collection) {
                            for (int i = 0 ; i < length ; ++i) {
                                collection.add(Array.get(value, i));
                            }
                        }
                    };
                } else {
                    return new Collector() {
                        public void collect(Collection collection) {
                            collection.add(value);
                        }
                    };
                }
            } catch (Exception shouldNeverHappen) {
                throw new Error(shouldNeverHappen);
            }
        }
    }
}
