/*
 * Copyright 2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.callbackparams.support;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import org.apache.bcel.Constants;
import org.apache.bcel.classfile.Attribute;
import org.apache.bcel.classfile.Code;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.classfile.Unknown;
import org.apache.bcel.generic.ALOAD;
import org.apache.bcel.generic.ClassGen;
import org.apache.bcel.generic.ConstantPoolGen;
import org.apache.bcel.generic.DUP;
import org.apache.bcel.generic.DUP2;
import org.apache.bcel.generic.DUP2_X1;
import org.apache.bcel.generic.DUP2_X2;
import org.apache.bcel.generic.DUP_X1;
import org.apache.bcel.generic.DUP_X2;
import org.apache.bcel.generic.EmptyVisitor;
import org.apache.bcel.generic.INVOKESPECIAL;
import org.apache.bcel.generic.Instruction;
import org.apache.bcel.generic.InstructionConstants;
import org.apache.bcel.generic.InstructionFactory;
import org.apache.bcel.generic.InstructionHandle;
import org.apache.bcel.generic.InstructionList;
import org.apache.bcel.generic.MethodGen;
import org.apache.bcel.generic.SWAP;
import org.apache.bcel.generic.Type;
import org.apache.bcel.generic.Visitor;
import org.apache.bcel.util.ClassLoaderRepository;

/**
 * A builder class for modifying class byte-code by staringt-off with an exising
 * template class. It comes with a few convience methods that can make some
 * common generical byte-code modifications.
 * One of the main features is to enable constant-time lookup of methods
 * through hashing - which might not be that important for most cases but could
 * prove useful once someone comes along with some generated class containing
 * thousands of methods :)
 *
 * @author Henrik Kaipe
 */
public class ClassBytecodeBuilder {

    /**
     * Is used for sending a "complete" exit-code from inside a visitor.
     */
    private static class VisitorTaskPerformed extends RuntimeException {
        private final static VisitorTaskPerformed instance =
                new VisitorTaskPerformed();
    }

    private final ClassGen cg;

    /**
     * Maps MethodHashKey instances to their respective bcel-Method instances.
     */
    private final Map methodMap = new HashMap();

    private ClassBytecodeBuilder(final ClassGen cg) {
        this.cg = cg;
        final Method[] methods = cg.getMethods();
        for (int i = 0 ; i < methods.length ; ++i) {
            final Method m = methods[i];
            methodMap.put(MethodHashKey.getHashKey(cg, m), m);
        }
    }

    public static ClassBytecodeBuilder newInstance(Class templateClass) {
        ClassLoaderRepository loaderRepo =
                new ClassLoaderRepository(templateClass.getClassLoader());
        try {
            return new ClassBytecodeBuilder(new ClassGen(
                    loaderRepo.loadClass(templateClass)));
        } catch (ClassNotFoundException ex) {
            throw new Error(ex);
        }
    }

    public InstructionFactory getInstructionFactory() {
        return new InstructionFactory(cg);
    }

    private static void filterUnknownAttributes(Code codeAttribute) {
        final Attribute[] subAttrs = codeAttribute.getAttributes();
        List list = new ArrayList(subAttrs.length);
        for (int i = 0; i < subAttrs.length; ++i) {
            if (false == subAttrs[i] instanceof Unknown) {
                list.add(subAttrs[i]);
            }
        }
        if (subAttrs.length != list.size()) {
            codeAttribute.setAttributes(
                    (Attribute[]) list.toArray(new Attribute[list.size()]));
        }
    }

    private static Method filterUnknownCodeAttributes(Method m) {
        Attribute[] attributes = m.getAttributes();
        for (int i = 0; i < attributes.length; ++i) {
            if (attributes[i] instanceof Code) {
                filterUnknownAttributes((Code) attributes[i]);
            }
        }
        return m;
    }

    public void prependMethod(Method m, InstructionList instructionList) {
        MethodGen mg = new MethodGen(
                m, cg.getClassName(), cg.getConstantPool());
        mg.getInstructionList().insert(instructionList);
        mg.setMaxStack();
        replaceMethod(filterUnknownCodeAttributes(mg.getMethod()));
    }

    public Method getMethod(java.lang.reflect.Method m) {
        return (Method) methodMap.get(MethodHashKey.getHashKey(m));
    }

    public ClassBytecodeBuilder replaceMethod(Method newMethod) {
        MethodHashKey hashkey = MethodHashKey.getHashKey(cg, newMethod);
        Method oldMethod = (Method) methodMap.get(hashkey);
        if (null == oldMethod) {
            throw new NullPointerException(
                    "There is no equivalent method to replace - " + newMethod);
        }
        cg.replaceMethod(oldMethod, newMethod);
        methodMap.put(hashkey, newMethod);
        return this;
    }

    public ClassBytecodeBuilder addMethod(Method m) {
        cg.addMethod(m);
        methodMap.put(MethodHashKey.getHashKey(cg, m), m);
        return this;
    }

    public byte[] getByteCode() {
        return cg.getJavaClass().getBytes();
    }

    public void setPublic() {
        cg.setModifiers(cg.getModifiers() | Constants.ACC_PUBLIC
                & ~Constants.ACC_PRIVATE & ~Constants.ACC_PROTECTED);
    }

    public void setNewSuperClass(String newSuperClassName) {
        for (Iterator i = methodMap.values().iterator() ; i.hasNext() ;) {
            final Method m = (Method) i.next();
            if ("<init>".equals(m.getName())) {
                modifyConstructor(m, newSuperClassName);
            }
        }
        cg.setSuperclassName(newSuperClassName);
    }

    public void setupTransparentConstructors(
            List constructorParamTypes, String newSuperClassName) {
        Method defaultConstructor = findDefaultConstructor();
        List instructionsToCopy = listInstructionsToKeep(defaultConstructor);

        /* Ditch the construction - it will be replaced by new ones. */
        removeMethod(defaultConstructor);

        for (Iterator i = constructorParamTypes.iterator() ; i.hasNext() ;) {
            final Type[] paramTypes = (Type[]) i.next();
            addNewTransparentConstructor(
                    paramTypes, newSuperClassName, instructionsToCopy);
        }
        cg.setSuperclassName(newSuperClassName);
    }

    private void addNewTransparentConstructor(final Type[] paramTypes,
            String newSuperClassName, List instructionsToCopy) {
        InstructionList il = new InstructionList();
        MethodGen mg = new MethodGen(Constants.ACC_PUBLIC, Type.VOID,
                paramTypes, null, "<init>", cg.getClassName(), il,
                cg.getConstantPool());

        il.append(InstructionConstants.ALOAD_0);
        for (int i = 0 ; i < paramTypes.length ; ++i) {
            il.append(InstructionFactory.createLoad(paramTypes[i], i + 1));
        }
        il.append(getInstructionFactory().createInvoke(newSuperClassName,
                "<init>", Type.VOID, paramTypes, Constants.INVOKESPECIAL));
        for (Iterator i = instructionsToCopy.iterator() ; i.hasNext() ;) {
            il.append((Instruction) i.next());
        }
        mg.setMaxStack();
        addMethod(mg.getMethod());
    }

    /**
     * Assumes there is only one constructor for this class.
     */
    private Method findDefaultConstructor() {
        for (Iterator i = methodMap.entrySet().iterator() ; i.hasNext() ;) {
            final Map.Entry entry = (Entry) i.next();
            final MethodHashKey key = (MethodHashKey) entry.getKey();
            if ("<init>".equals(key.methodName)) {
                return (Method) entry.getValue();
            }
        }
        throw new Error("No constructor was found");
    }

    private void modifyConstructor(
            Method method, final String newSuperClassName) {

        final String className = cg.getClassName();
        final Type thisType = Type.getType('L' + className + ';');
        final ConstantPoolGen cpg = cg.getConstantPool();
        final MethodGen mg = new MethodGen(method, className, cpg);

        new org.apache.bcel.generic.EmptyVisitor() {

            int stackDistanceToA0 = Integer.MAX_VALUE/2; // Initial dummy-value
            InstructionHandle ih;
            boolean pending = true;

//            @Override
            public void visitALOAD(ALOAD obj) {
                if (0 == obj.getIndex()) {
                    stackDistanceToA0 = 0;
                    throw VisitorTaskPerformed.instance;
                }
            }

            /** Used for handling the duplicating stack-operations DUP... */
            void donotModifyStackDistanceLessThan(int sizeOfUnchangedStackTop) {
                if (stackDistanceToA0 < sizeOfUnchangedStackTop) {
                    throw VisitorTaskPerformed.instance;
                }
            }

//            @Override
            public void visitDUP(DUP obj) {
                donotModifyStackDistanceLessThan(1);
            }

//            @Override
            public void visitDUP2(DUP2 obj) {
                donotModifyStackDistanceLessThan(2);
            }

//            @Override
            public void visitDUP2_X1(DUP2_X1 obj) {
                donotModifyStackDistanceLessThan(3);
            }

//            @Override
            public void visitDUP2_X2(DUP2_X2 obj) {
                donotModifyStackDistanceLessThan(4);
            }

//            @Override
            public void visitDUP_X1(DUP_X1 obj) {
                donotModifyStackDistanceLessThan(2);
            }

//            @Override
            public void visitDUP_X2(DUP_X2 obj) {
                donotModifyStackDistanceLessThan(3);
            }

//            @Override
            public void visitSWAP(SWAP obj) {
                if (stackDistanceToA0 < 2) {
                    stackDistanceToA0 = 1 - stackDistanceToA0;
                }
                throw VisitorTaskPerformed.instance;
            }

//            @Override
            public void visitINVOKESPECIAL(INVOKESPECIAL obj) {
//                System.out.println("Invoke special on " + className);
//                System.out.println("Distance to A0 :" + stackDistanceToA0);
//                System.out.println("Stack to consume: " + obj.consumeStack(cpg));
//                System.out.println("Instruction name: " + obj.getName(cpg));

                if (stackDistanceToA0 < obj.consumeStack(cpg)) {
                    assert "<init>".equals(obj.getName(cpg));
                    if (false == thisType.equals(obj.getReferenceType(cpg))) {
                        ih.setInstruction(getInstructionFactory().createInvoke(
                                newSuperClassName,
                                "<init>",
                                Type.VOID,
                                obj.getArgumentTypes(cpg),
                                Constants.INVOKESPECIAL));
                        mg.setMaxStack();
                        replaceMethod(mg.getMethod());
                    }
                    pending = false;
                    throw VisitorTaskPerformed.instance;
                }
            }

            void replaceSuperConstructorInvocation() {
                for (ih = mg.getInstructionList().getStart()
                        ; pending ; ih = ih.getNext()) {
                    try {
                        ih.accept(this);

                        Instruction instr = ih.getInstruction();
                        int produceStack = instr.produceStack(cpg);
                        if (0 < produceStack) {
                            stackDistanceToA0 += produceStack;
                        } else {
                            int consumeStack= instr.consumeStack(cpg);
                            if (0 < consumeStack) {
                                stackDistanceToA0 -= consumeStack;
                            }
                        }
                    } catch (VisitorTaskPerformed x) {
                        continue;
                    }
                }
            }

            /**
             * Initiator (An anonymous class cannot have a constructor)
             */
            {replaceSuperConstructorInvocation();}
        };
    }

    private List listInstructionsToKeep(Method defaultConstructor) {
        List instructionsToKeepInNewConstructors = new ArrayList();
        InstructionList il = new MethodGen(
                defaultConstructor, cg.getClassName(), cg.getConstantPool()).getInstructionList();
        Visitor superConstructorInvocationDetector = new EmptyVisitor() {
//            @Override
            public void visitINVOKESPECIAL(INVOKESPECIAL obj) {
                throw VisitorTaskPerformed.instance;
            }
        };
        Iterator i = il.iterator();
        try {
            while (i.hasNext()) {
                final InstructionHandle ih = (InstructionHandle) i.next();
                ih.accept(superConstructorInvocationDetector);
            }
            throw new Error("Found no super() invocation in instruction list: " + il);
        } catch (VisitorTaskPerformed x) {
            while (i.hasNext()) {
                final InstructionHandle ih = (InstructionHandle) i.next();
                instructionsToKeepInNewConstructors.add(ih.getInstruction());
            }
        }
        return instructionsToKeepInNewConstructors;
    }

    private void removeMethod(Method methodToRemove) {
        cg.removeMethod(methodToRemove);
        for (Iterator i = methodMap.entrySet().iterator() ; i.hasNext() ;) {
            Map.Entry entry = (Entry) i.next();
            if (methodToRemove == entry.getValue()) {
                i.remove();
                break;
            }
        }
    }
}
